print("import: 'intake'")
import intake

print("import: 'intake.auth'")
import intake.auth

print("import: 'intake.auth.tests'")
import intake.auth.tests

print("import: 'intake.catalog'")
import intake.catalog

print("import: 'intake.catalog.tests'")
import intake.catalog.tests

print("import: 'intake.cli'")
import intake.cli

print("import: 'intake.cli.client'")
import intake.cli.client

print("import: 'intake.cli.client.tests'")
import intake.cli.client.tests

print("import: 'intake.cli.server'")
import intake.cli.server

print("import: 'intake.cli.server.tests'")
import intake.cli.server.tests

print("import: 'intake.container'")
import intake.container

print("import: 'intake.gui'")
import intake.gui

print("import: 'intake.source'")
import intake.source

print("import: 'intake.source.tests'")
import intake.source.tests

print("import: 'intake.tests'")
import intake.tests

